function [ suggest best_acc best_sv acc ] = grid_svc ( svm, w, c, lcr, lgr, epsilon, n_fold, lambda, big_gamma, also_plot )
% 
%   function [ suggest best_acc best_sv ] = grid_svc ( svm, w, c, lcr, lgr, epsilon, n_fold, lambda, big_gamma, also_plot )
%       (n-fold cross-validation version, with a grid search rather than simulated annealing)
% 
% 
% Copyright (C) 2006, Matthew D. Boardman and Thomas P. Trappenberg, Dalhousie University, Nova Scotia, Canada
% 
%     This program is free software; you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation; either version 2 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License along
%     with this program; if not, write to the Free Software Foundation, Inc.,
%     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
% 

    %--------------------------------------------------------------
    % set defaults
    %--------------------------------------------------------------

    if nargin < 10
        also_plot=0;        % default to no plot
    end;

    if nargin < 9
        big_gamma=1/2;      % default to square root of percentage of support vectors
    end;
    
    if nargin < 8
        lambda=1;           % default to equal weight for accuracy and support vector ratio
    end;

    if nargin < 7
        n_fold = 5;         % default N-fold cross-validation
    end;

    if nargin < 6
        epsilon = 0.001;    % default to epsilon=0.001 
    end;

    if nargin < 5
        lgr = [ -15.5:1.0:5.5 ];  % wide range of log(g) values
    end;

    if nargin < 4
        lcr = [ -5.5:1.0:15.5 ];   % wide range of log(c) values
    end;

% be well-behaved in a crash:  set output variables to some value
        suggest=0;
        best_acc=0;
        best_sv=0;

    if nargin < 3

        fprintf('\nUsage:\n\n   [ suggest best_acc best_sv ] = grid_svc ( svm, w, c, lcr, lgr, epsilon, n_fold, lambda, big_gamma, also_plot )\n\n');

        fprintf('    Optimize cost and gamma parameters for support vector classifier with RBF kernel.\n');
        fprintf('    Uses simulated annealing, extrinsic regularization and intensity-weighted centre of mass.\n');
        fprintf('    This version uses N-fold cross-validation.\n\n');
        
        fprintf('These parameters must be specified:\n');
        fprintf('        svm  = 1 for LIBSVM or 2 for SVM-Light\n');
        fprintf('        w    = a matrix of training data (rows = training samples, columns = input dimensions)\n');
        fprintf('        c    = a vector of training labels (rows = training samples)\n');
        fprintf('\nThese parameters are optional:\n');
        fprintf('        lcr  = log cost range  (default is [ -5.5:1.0:15.5] )\n');
        fprintf('        lgr  = log gamma range (default is [-15.5:1.0: 5.5] )\n');
        fprintf('    epsilon  = SVM stopping criteria (default is 0.001)\n');
        fprintf('     n_fold  = number of data partions in N-fold cross-validation (default is 5)\n');
        fprintf('     lambda  = weight of support vector ratio in score (default is 1 = equal weight)\n');
        fprintf('   big_gamma = non-linearity for support vector ratio (default is .5 = sqrt, 1 is linear, 2 is square etc.)\n');
        fprintf('   also_plot = 1 to plot the simulated annealing path as you go (default is no)\n\n');

        fprintf('The parameter you will most want to play with is dt, which controls the length of the cooling cycle.\n');
        fprintf('A value of dt=0.1 gives 440 evaluations, whereas dt=0.01 gives 4590 evaluations.  More is better.\n');
        
        fprintf('Copyright (C) 2006, Matthew D. Boardman and Thomas P. Trappenberg, Dalhousie University, Nova Scotia, Canada.\n');
        fprintf('These MATLAB scripts come with ABSOLUTELY NO WARRANTY; for details, see GNU General Public License (license.txt).\n');
        fprintf('This is free software, and you are welcome to redistribute it under certain conditions; see license.txt for details.\n');
        fprintf('\n\n\n');
        
        return;
    end;
    
 
    %--------------------------------------------------------------
    % run the cross validation with a grid search
    %--------------------------------------------------------------

    % set initial point
    i=0;  
    j=0;  
    
    % guess that there will be 3000 accuracy points in the path (more allocated as we go)
    acc = zeros(3000,3);   
    true_acc = zeros(3000,3);
    true_sv = zeros(3000,3);
    
    current_acc = 0;
    best_acc = -1e6;
    best_true_acc = 0;
    best_true_sv = 0;
    best_num = 1;
        
    if  also_plot > 0
        figure(1); hold all;
        xlabel('log_e(cost)');
        ylabel('log_e(gamma)');
        axis ([ min(lcr) max(lcr) min(lgr) max(lgr) ]); 
        title('Grid Search');
    end;
    

    
    %-------------------------------MAIN LOOP -------------------------------------------

for new_log_gamma = lgr
    for new_log_cost = lcr
    
        % create parameter string for this run
        params = sprintf ( '-s 0 -t 2 -g %.8f -c %.8f -e %.8f', exp(new_log_gamma), exp(new_log_cost), epsilon );
        s_fold = sprintf ( ' -v %d', n_fold);
        
%-----------------------BEGIN TRIAL---------------------------

       % get cross-validation accuracy
        if svm == 1     % LIBSVM
            ac = svmtrain(c,w,cat(2,params,s_fold));
        else            % SVMLight
            ac = svmlearn(w,c,cat(2,params,s_fold));
        end;

       % train model (on all training data) to get number of SV
        if svm == 1     % LIBSVM
            model = svmtrain(c,w,params);
        else            % SVMLight
            model = svmlearn(w,c,params);
        end;

        new_acc = ac;
        
        new_sv = model.totalSV;        
        %new_sv = model{3};            % if above line causes an error, use this instead (old versions etc.)
    
        fprintf('Number of support vectors = %d\n', new_sv);
        
%----------------------END TRIAL--------------------------------        

        % calculate score 
        % note: lambda=1 for equal weight
        score = new_acc - lambda*100*((new_sv/size(w,1))^(big_gamma));
        
        fprintf ('#%4d: (Cost=exp(%3.1f),Gamma=exp(%3.1f)): Score=%.2f ', j+1, new_log_cost, new_log_gamma, score );

        % add to the list of known positions
        j=j+1;
        acc(j,1) = new_log_gamma;      
        acc(j,2) = new_log_cost;
        acc(j,3) = score; 
                
        % check best accuracy found so far
        if score > best_acc
            best_acc = score;
            clear best_list;
            best_num = 1;
            best_list(1,1) = new_log_cost;
            best_list(1,2) = new_log_gamma;
            best_true_acc = new_acc;
            best_true_sv = new_sv;
        elseif current_acc == best_acc
            best_num = best_num + 1;
            best_list ( best_num, 1 ) = log_cost;
            best_list ( best_num, 2 ) = log_gamma;
        end;

        fprintf('\t\t(Best:%.2f=%.2f+%.2f,lG=%.4f,lC=%.4f)\n', best_acc, best_true_acc, best_true_sv, best_list(1,2), best_list(1,1) );
        
    end;
end;


%-----------------------END MAIN LOOP--------------------------



%--------------------------------------------------------------
% calculate score-weighted centre of mass
%--------------------------------------------------------------

    fprintf('End of grid search.\n\nCalculating suggested parameters ... ');

    % get suggested point as minimum distance from (0,0)
    for i=1:best_num
        best_list(i,3) = sqrt(best_list(i,1)^2+best_list(i,2)^2);
    end;
    best_list=sortrows(best_list,3);
    suggest=best_list(1,1:2);
    
    % get overall best suggested point(s)
    clear best_list;
    best_num=0;
    psuggest=suggest;
    suggest=zeros(2,1);
    denom=0;

    acc=acc(1:j,:);   % cut off excess points if needed (since we guessed at the number of iterations)
   
    for i=1:j               
         if (sqrt((acc(i,1)-psuggest(1,2))^2+(acc(i,2)-psuggest(1,1))^2) < 1 )   % params are within log radius of 1?
            if ((best_acc > 0) & (acc(i,3) >= best_acc*.98)) | (best_acc < 0) & (acc(i,3) >= best_acc*1.02) % score is within 2% of best point?

                if best_acc > 0
                    suggest(1)=suggest(1)+(acc(i,3))*acc(i,2);  % cost
                    suggest(2)=suggest(2)+(acc(i,3))*acc(i,1);  % gamma
                    denom=denom+(acc(i,3));                     % score
                else    % reverse if scores are negative (not really ideal ... any ideas?)
                    suggest(1)=suggest(1)+(100+acc(i,3))*acc(i,2);  % cost
                    suggest(2)=suggest(2)+(100+acc(i,3))*acc(i,1);  % gamma
                    denom=denom+(100+acc(i,3));                     % score
                end;
                
                best_num=best_num+1;
                
                best_list(best_num,1)=acc(i,2);  % cost
                best_list(best_num,2)=acc(i,1);  % gamma
                best_list(best_num,3)=acc(i,3);  % score
            end;
         end;
    end;
  
    suggest=suggest/denom;
   
    % run SVM again with these parameters
    params = sprintf ( '-s 0 -t 2 -g %.8f -c %.8f -e %.8f', exp(suggest(2)), exp(suggest(1)), epsilon );
    s_fold = sprintf ( ' -v %d', n_fold);

    %---------- BEGIN TRIAL -----------

       % get cross-validation accuracy
        if svm == 1     % LIBSVM
            ac = svmtrain(c,w,cat(2,params,s_fold));
        else            % SVMLight
            ac = svmlearn(w,c,cat(2,params,s_fold));
        end;

        % train model (on all training data) to get number of SV
        if svm == 1     % LIBSVM
            model = svmtrain(c,w,params);
        else            % SVMLight
            model = svmlearn(w,c,params);
        end;

        new_acc = ac;

        new_sv = model.totalSV;        
        %new_sv = model{3};            % if above line causes an error, use this instead (old versions etc.)
    
    %---------- END TRIAL -----------

    suggest_acc = new_acc;
    suggest_sv = new_sv;
   
   % keep overall best points too, for display or comparison
    best_acc = best_true_acc;
    best_sv = best_true_sv;
    
    
%--------------------------------------------------------------
% plot a pretty grid
%--------------------------------------------------------------
    
    if also_plot > 0
        ac = reshape(acc(:,3), size(lcr,2), size(lgr,2))';
        hold all;
        plot3(suggest(1), suggest(2), 100,'^','MarkerEdgeColor','w','MarkerSize',12,'LineWidth',1,'MarkerFaceColor','w');
        plot3(psuggest(1,1), psuggest(1,2), 100,'s','MarkerEdgeColor','k','MarkerSize',12,'LineWidth',1);
        plot3(best_list(:,1), best_list(:,2), 100+0.*best_list(:,2),'o','MarkerEdgeColor','k','MarkerSize',4,'LineWidth',1); 
        surf(lcr,lgr,ac, 'linestyle','none'); 
        view(2); axis tight; grid off; colorbar;
    end;
    

%--------------------------------------------------------------
% detect failure, otherwise print results
%--------------------------------------------------------------

if best_acc == 0
    fprintf ( '\n\n!!! Resulting accuracy could not be measured (same over the entire range of values for cost and gamma).\n');
    return;
else
    fprintf ('\n\n--------------------\nDone! Top few points:\n');

    for i=1:best_num
        fprintf ('   Score = %.1f    Gamma = exp(%3.3f) = %.8f\t  Cost = exp(%3.3f) = %.8f\n', ...
            best_list(i,3), best_list(i,2), exp(best_list(i,2)), best_list(i,1), exp(best_list(i,1)) );
    end;
    
    fprintf('\nBest point found: \n');
    fprintf ('   Acc = %.1f  SV = %d/%d  Gamma = exp(%3.3f) = %.8f\t  Cost = exp(%3.3f) = %.8f\n', ...
        best_acc, best_sv, size(w,1), psuggest(1,2), exp(psuggest(1,2)), psuggest(1,1), exp(psuggest(1,1)) );
    
    fprintf('\nSuggested optimum point: (intensity-weighted centre of mass of best points)\n');
    fprintf ('   Acc = %.1f  SV = %d/%d  Gamma = exp(%3.3f) = %.8f\t  Cost = exp(%3.3f) = %.8f\n\n\n', ...
        suggest_acc, suggest_sv, size(w,1), suggest(2), exp(suggest(2)), suggest(1), exp(suggest(1)) );

end;

